/*
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2015 KYOCERA Corporation
 * (C) 2017 KYOCERA Corporation
 * (C) 2019 KYOCERA Corporation
 */
/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#define pr_fmt(fmt)	"WLED %s: " fmt, __func__

#include <linux/module.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include <linux/gpio.h>
#include <mach/kc_board.h>
#include "kc_wled.h"

#define EN_HI() gpio_set_value(gpio, 1)
#define EN_LO() gpio_set_value(gpio, 0)

#define EN_SET_RISING_EDGE() \
	EN_LO(); \
	udelay(10); \
	EN_HI()

#define EN_SETUP() \
	EN_HI(); \
	usleep(10)

#define CAT4004B_CURRENT_ST_MAX CAT4004B_CURRENT_ST_31

typedef enum {
  CAT4004B_CURRENT_ST_0,	/*    0 mA */
  CAT4004B_CURRENT_ST_1,	/*  0.8 mA */
  CAT4004B_CURRENT_ST_2,	/*  1.6 mA */
  CAT4004B_CURRENT_ST_3,	/*  2.4 mA */
  CAT4004B_CURRENT_ST_4,	/*  3.2 mA */
  CAT4004B_CURRENT_ST_5,	/*  4.0 mA */
  CAT4004B_CURRENT_ST_6,	/*  4.8 mA */
  CAT4004B_CURRENT_ST_7,	/*  5.6 mA */
  CAT4004B_CURRENT_ST_8,	/*  6.5 mA */
  CAT4004B_CURRENT_ST_9,	/*  7.3 mA */
  CAT4004B_CURRENT_ST_10,	/*  8.1 mA */
  CAT4004B_CURRENT_ST_11,	/*  8.9 mA */
  CAT4004B_CURRENT_ST_12,	/*  9.7 mA */
  CAT4004B_CURRENT_ST_13,	/* 10.5 mA */
  CAT4004B_CURRENT_ST_14,	/* 11.3 mA */
  CAT4004B_CURRENT_ST_15,	/* 12.1 mA */
  CAT4004B_CURRENT_ST_16,	/* 12.9 mA */
  CAT4004B_CURRENT_ST_17,	/* 13.7 mA */
  CAT4004B_CURRENT_ST_18,	/* 14.5 mA */
  CAT4004B_CURRENT_ST_19,	/* 15.3 mA */
  CAT4004B_CURRENT_ST_20,	/* 16.1 mA */
  CAT4004B_CURRENT_ST_21,	/* 16.9 mA */
  CAT4004B_CURRENT_ST_22,	/* 17.7 mA */
  CAT4004B_CURRENT_ST_23,	/* 18.5 mA */
  CAT4004B_CURRENT_ST_24,	/* 19.4 mA */
  CAT4004B_CURRENT_ST_25,	/* 20.2 mA */
  CAT4004B_CURRENT_ST_26,	/* 21.0 mA */
  CAT4004B_CURRENT_ST_27,	/* 21.8 mA */
  CAT4004B_CURRENT_ST_28,	/* 22.6 mA */
  CAT4004B_CURRENT_ST_29,	/* 23.4 mA */
  CAT4004B_CURRENT_ST_30,	/* 24.2 mA */
  CAT4004B_CURRENT_ST_31,	/* 25.0 mA */
  CAT4004B_CURRENT_ST_NUM
} current_state_type;

static const current_state_type current_state_tbl[WLED_BRIGHTNESS_NUM] = {
	CAT4004B_CURRENT_ST_0,	// 0
	CAT4004B_CURRENT_ST_0,	// 1
	CAT4004B_CURRENT_ST_0,	// 2
	CAT4004B_CURRENT_ST_0,	// 3
	CAT4004B_CURRENT_ST_0,	// 4
	CAT4004B_CURRENT_ST_0,	// 5
	CAT4004B_CURRENT_ST_0,	// 6
	CAT4004B_CURRENT_ST_0,	// 7
	CAT4004B_CURRENT_ST_1,	// 8
	CAT4004B_CURRENT_ST_1,	// 9
	CAT4004B_CURRENT_ST_1,	// 10
	CAT4004B_CURRENT_ST_1,	// 11
	CAT4004B_CURRENT_ST_1,	// 12
	CAT4004B_CURRENT_ST_1,	// 13
	CAT4004B_CURRENT_ST_1,	// 14
	CAT4004B_CURRENT_ST_1,	// 15
	CAT4004B_CURRENT_ST_2,	// 16
	CAT4004B_CURRENT_ST_2,	// 17
	CAT4004B_CURRENT_ST_2,	// 18
	CAT4004B_CURRENT_ST_2,	// 19
	CAT4004B_CURRENT_ST_2,	// 20
	CAT4004B_CURRENT_ST_2,	// 21
	CAT4004B_CURRENT_ST_2,	// 22
	CAT4004B_CURRENT_ST_2,	// 23
	CAT4004B_CURRENT_ST_3,	// 24
	CAT4004B_CURRENT_ST_3,	// 25
	CAT4004B_CURRENT_ST_3,	// 26
	CAT4004B_CURRENT_ST_3,	// 27
	CAT4004B_CURRENT_ST_3,	// 28
	CAT4004B_CURRENT_ST_3,	// 29
	CAT4004B_CURRENT_ST_3,	// 30
	CAT4004B_CURRENT_ST_3,	// 31
	CAT4004B_CURRENT_ST_4,	// 32
	CAT4004B_CURRENT_ST_4,	// 33
	CAT4004B_CURRENT_ST_4,	// 34
	CAT4004B_CURRENT_ST_4,	// 35
	CAT4004B_CURRENT_ST_4,	// 36
	CAT4004B_CURRENT_ST_4,	// 37
	CAT4004B_CURRENT_ST_4,	// 38
	CAT4004B_CURRENT_ST_4,	// 39
	CAT4004B_CURRENT_ST_5,	// 40
	CAT4004B_CURRENT_ST_5,	// 41
	CAT4004B_CURRENT_ST_5,	// 42
	CAT4004B_CURRENT_ST_5,	// 43
	CAT4004B_CURRENT_ST_5,	// 44
	CAT4004B_CURRENT_ST_5,	// 45
	CAT4004B_CURRENT_ST_5,	// 46
	CAT4004B_CURRENT_ST_5,	// 47
	CAT4004B_CURRENT_ST_6,	// 48
	CAT4004B_CURRENT_ST_6,	// 49
	CAT4004B_CURRENT_ST_6,	// 50
	CAT4004B_CURRENT_ST_6,	// 51
	CAT4004B_CURRENT_ST_6,	// 52
	CAT4004B_CURRENT_ST_6,	// 53
	CAT4004B_CURRENT_ST_6,	// 54
	CAT4004B_CURRENT_ST_6,	// 55
	CAT4004B_CURRENT_ST_7,	// 56
	CAT4004B_CURRENT_ST_7,	// 57
	CAT4004B_CURRENT_ST_7,	// 58
	CAT4004B_CURRENT_ST_7,	// 59
	CAT4004B_CURRENT_ST_7,	// 60
	CAT4004B_CURRENT_ST_7,	// 61
	CAT4004B_CURRENT_ST_7,	// 62
	CAT4004B_CURRENT_ST_7,	// 63
	CAT4004B_CURRENT_ST_8,	// 64
	CAT4004B_CURRENT_ST_8,	// 65
	CAT4004B_CURRENT_ST_8,	// 66
	CAT4004B_CURRENT_ST_8,	// 67
	CAT4004B_CURRENT_ST_8,	// 68
	CAT4004B_CURRENT_ST_8,	// 69
	CAT4004B_CURRENT_ST_8,	// 70
	CAT4004B_CURRENT_ST_8,	// 71
	CAT4004B_CURRENT_ST_9,	// 72
	CAT4004B_CURRENT_ST_9,	// 73
	CAT4004B_CURRENT_ST_9,	// 74
	CAT4004B_CURRENT_ST_9,	// 75
	CAT4004B_CURRENT_ST_9,	// 76
	CAT4004B_CURRENT_ST_9,	// 77
	CAT4004B_CURRENT_ST_9,	// 78
	CAT4004B_CURRENT_ST_9,	// 79
	CAT4004B_CURRENT_ST_10,	// 80
	CAT4004B_CURRENT_ST_10,	// 81
	CAT4004B_CURRENT_ST_10,	// 82
	CAT4004B_CURRENT_ST_10,	// 83
	CAT4004B_CURRENT_ST_10,	// 84
	CAT4004B_CURRENT_ST_10,	// 85
	CAT4004B_CURRENT_ST_10,	// 86
	CAT4004B_CURRENT_ST_10,	// 87
	CAT4004B_CURRENT_ST_11,	// 88
	CAT4004B_CURRENT_ST_11,	// 89
	CAT4004B_CURRENT_ST_11,	// 90
	CAT4004B_CURRENT_ST_11,	// 91
	CAT4004B_CURRENT_ST_11,	// 92
	CAT4004B_CURRENT_ST_11,	// 93
	CAT4004B_CURRENT_ST_11,	// 94
	CAT4004B_CURRENT_ST_11,	// 95
	CAT4004B_CURRENT_ST_12,	// 96
	CAT4004B_CURRENT_ST_12,	// 97
	CAT4004B_CURRENT_ST_12,	// 98
	CAT4004B_CURRENT_ST_12,	// 99
	CAT4004B_CURRENT_ST_12,	// 100
	CAT4004B_CURRENT_ST_12,	// 101
	CAT4004B_CURRENT_ST_12,	// 102
	CAT4004B_CURRENT_ST_12,	// 103
	CAT4004B_CURRENT_ST_13,	// 104
	CAT4004B_CURRENT_ST_13,	// 105
	CAT4004B_CURRENT_ST_13,	// 106
	CAT4004B_CURRENT_ST_13,	// 107
	CAT4004B_CURRENT_ST_13,	// 108
	CAT4004B_CURRENT_ST_13,	// 109
	CAT4004B_CURRENT_ST_13,	// 110
	CAT4004B_CURRENT_ST_13,	// 111
	CAT4004B_CURRENT_ST_14,	// 112
	CAT4004B_CURRENT_ST_14,	// 113
	CAT4004B_CURRENT_ST_14,	// 114
	CAT4004B_CURRENT_ST_14,	// 115
	CAT4004B_CURRENT_ST_14,	// 116
	CAT4004B_CURRENT_ST_14,	// 117
	CAT4004B_CURRENT_ST_14,	// 118
	CAT4004B_CURRENT_ST_14,	// 119
	CAT4004B_CURRENT_ST_15,	// 120
	CAT4004B_CURRENT_ST_15,	// 121
	CAT4004B_CURRENT_ST_15,	// 122
	CAT4004B_CURRENT_ST_15,	// 123
	CAT4004B_CURRENT_ST_15,	// 124
	CAT4004B_CURRENT_ST_15,	// 125
	CAT4004B_CURRENT_ST_15,	// 126
	CAT4004B_CURRENT_ST_15,	// 127
	CAT4004B_CURRENT_ST_16,	// 128
	CAT4004B_CURRENT_ST_16,	// 129
	CAT4004B_CURRENT_ST_16,	// 130
	CAT4004B_CURRENT_ST_16,	// 131
	CAT4004B_CURRENT_ST_16,	// 132
	CAT4004B_CURRENT_ST_16,	// 133
	CAT4004B_CURRENT_ST_16,	// 134
	CAT4004B_CURRENT_ST_16,	// 135
	CAT4004B_CURRENT_ST_17,	// 136
	CAT4004B_CURRENT_ST_17,	// 137
	CAT4004B_CURRENT_ST_17,	// 138
	CAT4004B_CURRENT_ST_17,	// 139
	CAT4004B_CURRENT_ST_17,	// 140
	CAT4004B_CURRENT_ST_17,	// 141
	CAT4004B_CURRENT_ST_17,	// 142
	CAT4004B_CURRENT_ST_17,	// 143
	CAT4004B_CURRENT_ST_18,	// 144
	CAT4004B_CURRENT_ST_18,	// 145
	CAT4004B_CURRENT_ST_18,	// 146
	CAT4004B_CURRENT_ST_18,	// 147
	CAT4004B_CURRENT_ST_18,	// 148
	CAT4004B_CURRENT_ST_18,	// 149
	CAT4004B_CURRENT_ST_18,	// 150
	CAT4004B_CURRENT_ST_18,	// 151
	CAT4004B_CURRENT_ST_19,	// 152
	CAT4004B_CURRENT_ST_19,	// 153
	CAT4004B_CURRENT_ST_19,	// 154
	CAT4004B_CURRENT_ST_19,	// 155
	CAT4004B_CURRENT_ST_19,	// 156
	CAT4004B_CURRENT_ST_19,	// 157
	CAT4004B_CURRENT_ST_19,	// 158
	CAT4004B_CURRENT_ST_19,	// 159
	CAT4004B_CURRENT_ST_20,	// 160
	CAT4004B_CURRENT_ST_20,	// 161
	CAT4004B_CURRENT_ST_20,	// 162
	CAT4004B_CURRENT_ST_20,	// 163
	CAT4004B_CURRENT_ST_20,	// 164
	CAT4004B_CURRENT_ST_20,	// 165
	CAT4004B_CURRENT_ST_20,	// 166
	CAT4004B_CURRENT_ST_20,	// 167
	CAT4004B_CURRENT_ST_21,	// 168
	CAT4004B_CURRENT_ST_21,	// 169
	CAT4004B_CURRENT_ST_21,	// 170
	CAT4004B_CURRENT_ST_21,	// 171
	CAT4004B_CURRENT_ST_21,	// 172
	CAT4004B_CURRENT_ST_21,	// 173
	CAT4004B_CURRENT_ST_21,	// 174
	CAT4004B_CURRENT_ST_21,	// 175
	CAT4004B_CURRENT_ST_22,	// 176
	CAT4004B_CURRENT_ST_22,	// 177
	CAT4004B_CURRENT_ST_22,	// 178
	CAT4004B_CURRENT_ST_22,	// 179
	CAT4004B_CURRENT_ST_22,	// 180
	CAT4004B_CURRENT_ST_22,	// 181
	CAT4004B_CURRENT_ST_22,	// 182
	CAT4004B_CURRENT_ST_22,	// 183
	CAT4004B_CURRENT_ST_23,	// 184
	CAT4004B_CURRENT_ST_23,	// 185
	CAT4004B_CURRENT_ST_23,	// 186
	CAT4004B_CURRENT_ST_23,	// 187
	CAT4004B_CURRENT_ST_23,	// 188
	CAT4004B_CURRENT_ST_23,	// 189
	CAT4004B_CURRENT_ST_23,	// 190
	CAT4004B_CURRENT_ST_23,	// 191
	CAT4004B_CURRENT_ST_24,	// 192
	CAT4004B_CURRENT_ST_24,	// 193
	CAT4004B_CURRENT_ST_24,	// 194
	CAT4004B_CURRENT_ST_24,	// 195
	CAT4004B_CURRENT_ST_24,	// 196
	CAT4004B_CURRENT_ST_24,	// 197
	CAT4004B_CURRENT_ST_24,	// 198
	CAT4004B_CURRENT_ST_24,	// 199
	CAT4004B_CURRENT_ST_25,	// 200
	CAT4004B_CURRENT_ST_25,	// 201
	CAT4004B_CURRENT_ST_25,	// 202
	CAT4004B_CURRENT_ST_25,	// 203
	CAT4004B_CURRENT_ST_25,	// 204
	CAT4004B_CURRENT_ST_25,	// 205
	CAT4004B_CURRENT_ST_25,	// 206
	CAT4004B_CURRENT_ST_25,	// 207
	CAT4004B_CURRENT_ST_26,	// 208
	CAT4004B_CURRENT_ST_26,	// 209
	CAT4004B_CURRENT_ST_26,	// 210
	CAT4004B_CURRENT_ST_26,	// 211
	CAT4004B_CURRENT_ST_26,	// 212
	CAT4004B_CURRENT_ST_26,	// 213
	CAT4004B_CURRENT_ST_26,	// 214
	CAT4004B_CURRENT_ST_26,	// 215
	CAT4004B_CURRENT_ST_27,	// 216
	CAT4004B_CURRENT_ST_27,	// 217
	CAT4004B_CURRENT_ST_27,	// 218
	CAT4004B_CURRENT_ST_27,	// 219
	CAT4004B_CURRENT_ST_27,	// 220
	CAT4004B_CURRENT_ST_27,	// 221
	CAT4004B_CURRENT_ST_27,	// 222
	CAT4004B_CURRENT_ST_27,	// 223
	CAT4004B_CURRENT_ST_28,	// 224
	CAT4004B_CURRENT_ST_28,	// 225
	CAT4004B_CURRENT_ST_28,	// 226
	CAT4004B_CURRENT_ST_28,	// 227
	CAT4004B_CURRENT_ST_28,	// 228
	CAT4004B_CURRENT_ST_28,	// 229
	CAT4004B_CURRENT_ST_28,	// 230
	CAT4004B_CURRENT_ST_28,	// 231
	CAT4004B_CURRENT_ST_28,	// 232
	CAT4004B_CURRENT_ST_28,	// 233
	CAT4004B_CURRENT_ST_28,	// 234
	CAT4004B_CURRENT_ST_28,	// 235
	CAT4004B_CURRENT_ST_28,	// 236
	CAT4004B_CURRENT_ST_28,	// 237
	CAT4004B_CURRENT_ST_28,	// 238
	CAT4004B_CURRENT_ST_28,	// 239
	CAT4004B_CURRENT_ST_28,	// 240
	CAT4004B_CURRENT_ST_28,	// 241
	CAT4004B_CURRENT_ST_28,	// 242
	CAT4004B_CURRENT_ST_28,	// 243
	CAT4004B_CURRENT_ST_28,	// 244
	CAT4004B_CURRENT_ST_28,	// 245
	CAT4004B_CURRENT_ST_28,	// 246
	CAT4004B_CURRENT_ST_28,	// 247
	CAT4004B_CURRENT_ST_28,	// 248
	CAT4004B_CURRENT_ST_28,	// 249
	CAT4004B_CURRENT_ST_28,	// 250
	CAT4004B_CURRENT_ST_28,	// 251
	CAT4004B_CURRENT_ST_28,	// 252
	CAT4004B_CURRENT_ST_28,	// 253
	CAT4004B_CURRENT_ST_28,	// 254
	CAT4004B_CURRENT_ST_28,	// 255
};

static const current_state_type current_state_tbl_2[WLED_BRIGHTNESS_NUM] = {
	CAT4004B_CURRENT_ST_0,	// 0
	CAT4004B_CURRENT_ST_0,	// 1
	CAT4004B_CURRENT_ST_0,	// 2
	CAT4004B_CURRENT_ST_0,	// 3
	CAT4004B_CURRENT_ST_0,	// 4
	CAT4004B_CURRENT_ST_0,	// 5
	CAT4004B_CURRENT_ST_0,	// 6
	CAT4004B_CURRENT_ST_0,	// 7
	CAT4004B_CURRENT_ST_1,	// 8
	CAT4004B_CURRENT_ST_1,	// 9
	CAT4004B_CURRENT_ST_1,	// 10
	CAT4004B_CURRENT_ST_1,	// 11
	CAT4004B_CURRENT_ST_1,	// 12
	CAT4004B_CURRENT_ST_1,	// 13
	CAT4004B_CURRENT_ST_1,	// 14
	CAT4004B_CURRENT_ST_1,	// 15
	CAT4004B_CURRENT_ST_2,	// 16
	CAT4004B_CURRENT_ST_2,	// 17
	CAT4004B_CURRENT_ST_2,	// 18
	CAT4004B_CURRENT_ST_2,	// 19
	CAT4004B_CURRENT_ST_2,	// 20
	CAT4004B_CURRENT_ST_2,	// 21
	CAT4004B_CURRENT_ST_2,	// 22
	CAT4004B_CURRENT_ST_2,	// 23
	CAT4004B_CURRENT_ST_3,	// 24
	CAT4004B_CURRENT_ST_3,	// 25
	CAT4004B_CURRENT_ST_3,	// 26
	CAT4004B_CURRENT_ST_3,	// 27
	CAT4004B_CURRENT_ST_3,	// 28
	CAT4004B_CURRENT_ST_3,	// 29
	CAT4004B_CURRENT_ST_3,	// 30
	CAT4004B_CURRENT_ST_3,	// 31
	CAT4004B_CURRENT_ST_4,	// 32
	CAT4004B_CURRENT_ST_4,	// 33
	CAT4004B_CURRENT_ST_4,	// 34
	CAT4004B_CURRENT_ST_4,	// 35
	CAT4004B_CURRENT_ST_4,	// 36
	CAT4004B_CURRENT_ST_4,	// 37
	CAT4004B_CURRENT_ST_4,	// 38
	CAT4004B_CURRENT_ST_4,	// 39
	CAT4004B_CURRENT_ST_5,	// 40
	CAT4004B_CURRENT_ST_5,	// 41
	CAT4004B_CURRENT_ST_5,	// 42
	CAT4004B_CURRENT_ST_5,	// 43
	CAT4004B_CURRENT_ST_5,	// 44
	CAT4004B_CURRENT_ST_5,	// 45
	CAT4004B_CURRENT_ST_5,	// 46
	CAT4004B_CURRENT_ST_5,	// 47
	CAT4004B_CURRENT_ST_6,	// 48
	CAT4004B_CURRENT_ST_6,	// 49
	CAT4004B_CURRENT_ST_6,	// 50
	CAT4004B_CURRENT_ST_6,	// 51
	CAT4004B_CURRENT_ST_6,	// 52
	CAT4004B_CURRENT_ST_6,	// 53
	CAT4004B_CURRENT_ST_6,	// 54
	CAT4004B_CURRENT_ST_6,	// 55
	CAT4004B_CURRENT_ST_7,	// 56
	CAT4004B_CURRENT_ST_7,	// 57
	CAT4004B_CURRENT_ST_7,	// 58
	CAT4004B_CURRENT_ST_7,	// 59
	CAT4004B_CURRENT_ST_7,	// 60
	CAT4004B_CURRENT_ST_7,	// 61
	CAT4004B_CURRENT_ST_7,	// 62
	CAT4004B_CURRENT_ST_7,	// 63
	CAT4004B_CURRENT_ST_8,	// 64
	CAT4004B_CURRENT_ST_8,	// 65
	CAT4004B_CURRENT_ST_8,	// 66
	CAT4004B_CURRENT_ST_8,	// 67
	CAT4004B_CURRENT_ST_8,	// 68
	CAT4004B_CURRENT_ST_8,	// 69
	CAT4004B_CURRENT_ST_8,	// 70
	CAT4004B_CURRENT_ST_8,	// 71
	CAT4004B_CURRENT_ST_9,	// 72
	CAT4004B_CURRENT_ST_9,	// 73
	CAT4004B_CURRENT_ST_9,	// 74
	CAT4004B_CURRENT_ST_9,	// 75
	CAT4004B_CURRENT_ST_9,	// 76
	CAT4004B_CURRENT_ST_9,	// 77
	CAT4004B_CURRENT_ST_9,	// 78
	CAT4004B_CURRENT_ST_9,	// 79
	CAT4004B_CURRENT_ST_10,	// 80
	CAT4004B_CURRENT_ST_10,	// 81
	CAT4004B_CURRENT_ST_10,	// 82
	CAT4004B_CURRENT_ST_10,	// 83
	CAT4004B_CURRENT_ST_10,	// 84
	CAT4004B_CURRENT_ST_10,	// 85
	CAT4004B_CURRENT_ST_10,	// 86
	CAT4004B_CURRENT_ST_10,	// 87
	CAT4004B_CURRENT_ST_11,	// 88
	CAT4004B_CURRENT_ST_11,	// 89
	CAT4004B_CURRENT_ST_11,	// 90
	CAT4004B_CURRENT_ST_11,	// 91
	CAT4004B_CURRENT_ST_11,	// 92
	CAT4004B_CURRENT_ST_11,	// 93
	CAT4004B_CURRENT_ST_11,	// 94
	CAT4004B_CURRENT_ST_11,	// 95
	CAT4004B_CURRENT_ST_11,	// 96
	CAT4004B_CURRENT_ST_11,	// 97
	CAT4004B_CURRENT_ST_11,	// 98
	CAT4004B_CURRENT_ST_11,	// 99
	CAT4004B_CURRENT_ST_11,	// 100
	CAT4004B_CURRENT_ST_11,	// 101
	CAT4004B_CURRENT_ST_11,	// 102
	CAT4004B_CURRENT_ST_11,	// 103
	CAT4004B_CURRENT_ST_11,	// 104
	CAT4004B_CURRENT_ST_11,	// 105
	CAT4004B_CURRENT_ST_11,	// 106
	CAT4004B_CURRENT_ST_11,	// 107
	CAT4004B_CURRENT_ST_11,	// 108
	CAT4004B_CURRENT_ST_11,	// 109
	CAT4004B_CURRENT_ST_11,	// 110
	CAT4004B_CURRENT_ST_11,	// 111
	CAT4004B_CURRENT_ST_12,	// 112
	CAT4004B_CURRENT_ST_12,	// 113
	CAT4004B_CURRENT_ST_12,	// 114
	CAT4004B_CURRENT_ST_12,	// 115
	CAT4004B_CURRENT_ST_12,	// 116
	CAT4004B_CURRENT_ST_12,	// 117
	CAT4004B_CURRENT_ST_12,	// 118
	CAT4004B_CURRENT_ST_12,	// 119
	CAT4004B_CURRENT_ST_12,	// 120
	CAT4004B_CURRENT_ST_12,	// 121
	CAT4004B_CURRENT_ST_12,	// 122
	CAT4004B_CURRENT_ST_12,	// 123
	CAT4004B_CURRENT_ST_12,	// 124
	CAT4004B_CURRENT_ST_12,	// 125
	CAT4004B_CURRENT_ST_12,	// 126
	CAT4004B_CURRENT_ST_12,	// 127
	CAT4004B_CURRENT_ST_13,	// 128
	CAT4004B_CURRENT_ST_13,	// 129
	CAT4004B_CURRENT_ST_13,	// 130
	CAT4004B_CURRENT_ST_13,	// 131
	CAT4004B_CURRENT_ST_13,	// 132
	CAT4004B_CURRENT_ST_13,	// 133
	CAT4004B_CURRENT_ST_13,	// 134
	CAT4004B_CURRENT_ST_13,	// 135
	CAT4004B_CURRENT_ST_14,	// 136
	CAT4004B_CURRENT_ST_14,	// 137
	CAT4004B_CURRENT_ST_14,	// 138
	CAT4004B_CURRENT_ST_14,	// 139
	CAT4004B_CURRENT_ST_14,	// 140
	CAT4004B_CURRENT_ST_14,	// 141
	CAT4004B_CURRENT_ST_14,	// 142
	CAT4004B_CURRENT_ST_14,	// 143
	CAT4004B_CURRENT_ST_18,	// 144
	CAT4004B_CURRENT_ST_18,	// 145
	CAT4004B_CURRENT_ST_18,	// 146
	CAT4004B_CURRENT_ST_18,	// 147
	CAT4004B_CURRENT_ST_18,	// 148
	CAT4004B_CURRENT_ST_18,	// 149
	CAT4004B_CURRENT_ST_18,	// 150
	CAT4004B_CURRENT_ST_18,	// 151
	CAT4004B_CURRENT_ST_19,	// 152
	CAT4004B_CURRENT_ST_19,	// 153
	CAT4004B_CURRENT_ST_19,	// 154
	CAT4004B_CURRENT_ST_19,	// 155
	CAT4004B_CURRENT_ST_19,	// 156
	CAT4004B_CURRENT_ST_19,	// 157
	CAT4004B_CURRENT_ST_19,	// 158
	CAT4004B_CURRENT_ST_19,	// 159
	CAT4004B_CURRENT_ST_20,	// 160
	CAT4004B_CURRENT_ST_20,	// 161
	CAT4004B_CURRENT_ST_20,	// 162
	CAT4004B_CURRENT_ST_20,	// 163
	CAT4004B_CURRENT_ST_20,	// 164
	CAT4004B_CURRENT_ST_20,	// 165
	CAT4004B_CURRENT_ST_20,	// 166
	CAT4004B_CURRENT_ST_20,	// 167
	CAT4004B_CURRENT_ST_21,	// 168
	CAT4004B_CURRENT_ST_21,	// 169
	CAT4004B_CURRENT_ST_21,	// 170
	CAT4004B_CURRENT_ST_21,	// 171
	CAT4004B_CURRENT_ST_21,	// 172
	CAT4004B_CURRENT_ST_21,	// 173
	CAT4004B_CURRENT_ST_21,	// 174
	CAT4004B_CURRENT_ST_21,	// 175
	CAT4004B_CURRENT_ST_22,	// 176
	CAT4004B_CURRENT_ST_22,	// 177
	CAT4004B_CURRENT_ST_22,	// 178
	CAT4004B_CURRENT_ST_22,	// 179
	CAT4004B_CURRENT_ST_22,	// 180
	CAT4004B_CURRENT_ST_22,	// 181
	CAT4004B_CURRENT_ST_22,	// 182
	CAT4004B_CURRENT_ST_22,	// 183
	CAT4004B_CURRENT_ST_23,	// 184
	CAT4004B_CURRENT_ST_23,	// 185
	CAT4004B_CURRENT_ST_23,	// 186
	CAT4004B_CURRENT_ST_23,	// 187
	CAT4004B_CURRENT_ST_23,	// 188
	CAT4004B_CURRENT_ST_23,	// 189
	CAT4004B_CURRENT_ST_23,	// 190
	CAT4004B_CURRENT_ST_23,	// 191
	CAT4004B_CURRENT_ST_24,	// 192
	CAT4004B_CURRENT_ST_24,	// 193
	CAT4004B_CURRENT_ST_24,	// 194
	CAT4004B_CURRENT_ST_24,	// 195
	CAT4004B_CURRENT_ST_24,	// 196
	CAT4004B_CURRENT_ST_24,	// 197
	CAT4004B_CURRENT_ST_24,	// 198
	CAT4004B_CURRENT_ST_24,	// 199
	CAT4004B_CURRENT_ST_25,	// 200
	CAT4004B_CURRENT_ST_25,	// 201
	CAT4004B_CURRENT_ST_25,	// 202
	CAT4004B_CURRENT_ST_25,	// 203
	CAT4004B_CURRENT_ST_25,	// 204
	CAT4004B_CURRENT_ST_25,	// 205
	CAT4004B_CURRENT_ST_25,	// 206
	CAT4004B_CURRENT_ST_25,	// 207
	CAT4004B_CURRENT_ST_26,	// 208
	CAT4004B_CURRENT_ST_26,	// 209
	CAT4004B_CURRENT_ST_26,	// 210
	CAT4004B_CURRENT_ST_26,	// 211
	CAT4004B_CURRENT_ST_26,	// 212
	CAT4004B_CURRENT_ST_26,	// 213
	CAT4004B_CURRENT_ST_26,	// 214
	CAT4004B_CURRENT_ST_26,	// 215
	CAT4004B_CURRENT_ST_27,	// 216
	CAT4004B_CURRENT_ST_27,	// 217
	CAT4004B_CURRENT_ST_27,	// 218
	CAT4004B_CURRENT_ST_27,	// 219
	CAT4004B_CURRENT_ST_27,	// 220
	CAT4004B_CURRENT_ST_27,	// 221
	CAT4004B_CURRENT_ST_27,	// 222
	CAT4004B_CURRENT_ST_27,	// 223
	CAT4004B_CURRENT_ST_28,	// 224
	CAT4004B_CURRENT_ST_28,	// 225
	CAT4004B_CURRENT_ST_28,	// 226
	CAT4004B_CURRENT_ST_28,	// 227
	CAT4004B_CURRENT_ST_28,	// 228
	CAT4004B_CURRENT_ST_28,	// 229
	CAT4004B_CURRENT_ST_28,	// 230
	CAT4004B_CURRENT_ST_28,	// 231
	CAT4004B_CURRENT_ST_28,	// 232
	CAT4004B_CURRENT_ST_28,	// 233
	CAT4004B_CURRENT_ST_28,	// 234
	CAT4004B_CURRENT_ST_28,	// 235
	CAT4004B_CURRENT_ST_28,	// 236
	CAT4004B_CURRENT_ST_28,	// 237
	CAT4004B_CURRENT_ST_28,	// 238
	CAT4004B_CURRENT_ST_28,	// 239
	CAT4004B_CURRENT_ST_28,	// 240
	CAT4004B_CURRENT_ST_28,	// 241
	CAT4004B_CURRENT_ST_28,	// 242
	CAT4004B_CURRENT_ST_28,	// 243
	CAT4004B_CURRENT_ST_28,	// 244
	CAT4004B_CURRENT_ST_28,	// 245
	CAT4004B_CURRENT_ST_28,	// 246
	CAT4004B_CURRENT_ST_28,	// 247
	CAT4004B_CURRENT_ST_28,	// 248
	CAT4004B_CURRENT_ST_28,	// 249
	CAT4004B_CURRENT_ST_28,	// 250
	CAT4004B_CURRENT_ST_28,	// 251
	CAT4004B_CURRENT_ST_28,	// 252
	CAT4004B_CURRENT_ST_28,	// 253
	CAT4004B_CURRENT_ST_28,	// 254
	CAT4004B_CURRENT_ST_28,	// 255
};

static DEFINE_SPINLOCK(cat4004b_lock);

static int current_state = CAT4004B_CURRENT_ST_10;
static int wled_hw_id = OEM_BOARD_FAIL_TYPE;
static unsigned long wled_stat_aboot = WLED_BRIGHTNESS_DEFAULT;

static int cat4004b_get_hwid(void)
{
	if (wled_hw_id == OEM_BOARD_FAIL_TYPE) {
		wled_hw_id = OEM_get_board();
		pr_notice("%s: hw id = %d\n", __func__, wled_hw_id);
	}

	return wled_hw_id;
}

static int __init cat4004b_set_stat_aboot(char *buf)
{
	int ret;

	pr_debug("[IN]\n");
	ret = kstrtoul(buf, 0, &wled_stat_aboot);
	if (ret) {
		pr_err("[OUT] cat4004b_set_stat_aboot param error\n");
	}
	pr_notice("[OUT] wled_stat_aboot=0x%lx\n", wled_stat_aboot);
	return 0;
}
early_param("led_stat", cat4004b_set_stat_aboot);

void cat4004b_init(void)
{
	int hw_id;

	hw_id = cat4004b_get_hwid();
	if (hw_id == OEM_BOARD_FAIL_TYPE) {
		current_state = current_state_tbl[wled_stat_aboot];
	} else if (hw_id < OEM_BOARD_MP2_TYPE) {
		current_state = current_state_tbl[wled_stat_aboot];
	} else {
		current_state = current_state_tbl_2[wled_stat_aboot];
	}
	pr_notice("[OUT] cat4004b_init=%d\n", current_state);
}

void cat4004b_work(int level, enum kc_wled_table tbl, int gpio)
{
	int new_current_state;
	int shift_cnt = 0;
	unsigned long flags;
	int hw_id;

	hw_id = cat4004b_get_hwid();
	if (hw_id == OEM_BOARD_FAIL_TYPE) {
		new_current_state = current_state_tbl[level];
	} else if (hw_id < OEM_BOARD_MP2_TYPE) {
		new_current_state = current_state_tbl[level];
	} else {
		new_current_state = current_state_tbl_2[level];
	}

	pr_debug("%s: level = %d new_current_state=%d\n", __func__, level, new_current_state);

	if (current_state == new_current_state) {
		pr_debug("do not set because it is same value level:%d tbl:%d current_state:%d\n", level, tbl, current_state);
		return;
	}

	if (current_state < new_current_state) {
		shift_cnt = CAT4004B_CURRENT_ST_NUM - (new_current_state - current_state);

		if (current_state == 0) {
			EN_SETUP();
			shift_cnt--;
			pr_debug("shift_cnt-- to %d level:%d current_state:%d\n", shift_cnt, level, current_state);
		}
	} else {
		shift_cnt = current_state - new_current_state;
	}
	pr_debug("shift_cnt:%d\n", shift_cnt);

	spin_lock_irqsave(&cat4004b_lock, flags);
	for (; shift_cnt != 0; shift_cnt--) {
		EN_SET_RISING_EDGE();
		udelay(10);
	}
	spin_unlock_irqrestore(&cat4004b_lock, flags);

	if (new_current_state == 0) {
		EN_LO();
		pr_debug("CAT4004B OFF current_state:%d\n", current_state);
	}
	pr_debug("level:%d tbl:%d current_state old:%d new:%d\n", level, tbl, current_state, new_current_state);
	current_state = new_current_state;

	return;
}

